import os
import requests
import pandas as pd
import zipfile
import io
import numpy as np
import xarray as xr
import geopandas as gpd
from shapely.geometry import Point

TARGET_DIR = '/net/fs03/d1/gchossie/2020_aq/monitor_data/'
if TARGET_DIR == '':
    datadir = './raw_data/'
    if not os.path.isdir(datadir):
        os.mkdir(datadir)
else:
    datadir = TARGET_DIR + 'raw_data/'
    if not os.path.isdir(datadir):
        os.mkdir(datadir)

# Define a US state code to name conversion
us_state_abbrev = {
    'Alabama': 'AL',
    'Alaska': 'AK',
    'American Samoa': 'AS',
    'Arizona': 'AZ',
    'Arkansas': 'AR',
    'California': 'CA',
    'Colorado': 'CO',
    'Connecticut': 'CT',
    'Delaware': 'DE',
    'District of Columbia': 'DC',
    'Florida': 'FL',
    'Georgia': 'GA',
    'Guam': 'GU',
    'Hawaii': 'HI',
    'Idaho': 'ID',
    'Illinois': 'IL',
    'Indiana': 'IN',
    'Iowa': 'IA',
    'Kansas': 'KS',
    'Kentucky': 'KY',
    'Louisiana': 'LA',
    'Maine': 'ME',
    'Maryland': 'MD',
    'Massachusetts': 'MA',
    'Michigan': 'MI',
    'Minnesota': 'MN',
    'Mississippi': 'MS',
    'Missouri': 'MO',
    'Montana': 'MT',
    'Nebraska': 'NE',
    'Nevada': 'NV',
    'New Hampshire': 'NH',
    'New Jersey': 'NJ',
    'New Mexico': 'NM',
    'New York': 'NY',
    'North Carolina': 'NC',
    'North Dakota': 'ND',
    'Northern Mariana Islands':'MP',
    'Ohio': 'OH',
    'Oklahoma': 'OK',
    'Oregon': 'OR',
    'Pennsylvania': 'PA',
    'Puerto Rico': 'PR',
    'Rhode Island': 'RI',
    'South Carolina': 'SC',
    'South Dakota': 'SD',
    'Tennessee': 'TN',
    'Texas': 'TX',
    'Utah': 'UT',
    'Vermont': 'VT',
    'Virgin Islands': 'VI',
    'Virginia': 'VA',
    'Washington': 'WA',
    'West Virginia': 'WV',
    'Wisconsin': 'WI',
    'Wyoming': 'WY'
}
us_abbrev_state = {v: k for k,v in us_state_abbrev.items()}
all_regions = gpd.read_file('./ext_data/all_regions.geojson')

def us_monitor_utility() -> None:
    '''
    Download the most recent files with monitor information
    '''
    directory = datadir + 'us/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    url_monitors = 'https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/today/monitoring_site_locations.dat'
    monloc = pd.read_csv(url_monitors, sep='|', header=None,
                     encoding='ISO-8859-1')
    lst = ['AQSID', 'parametername', 'sitecode', 'sitename', 'status', 'agencyid', 'agencyname',
           'EPAregion', 'latitude', 'longitude', 'elevation', 'GMToffset', 'countrycode', 'MSAcode',
                                    'MSAname', 'statecode', 'statename', 'countycode', 'countyname']
    monloc = monloc.rename(columns={i: lst[i] for i in range(len(lst))})
    # Clean-up
    monloc = monloc.loc[monloc['status'] == 'Active'].reset_index(drop=True)
    monloc = monloc.drop(columns=[19,20,21,22])
    idx = (monloc['longitude'] >= -172.5).values * \
        (monloc['longitude'] <= -36).values
    monloc = monloc.loc[idx]
    monloc.reset_index(drop=False).to_feather(directory + 'us_monitoring_site_locations.feather')
    return None

def jp_monitor_utility() -> None:
    '''
    Download the most recent files with monitor information
    '''
    directory = datadir + 'japan/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    url = 'http://www.nies.go.jp/igreen/tm/2017/TM20170000.zip'
    resp = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(resp.content))
    z.extractall(directory)
    return None

def kr_monitor_utility() -> None:
    '''
    Download the most recent files with monitor information
    '''
    directory = datadir + 'korea/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    os.system(f'curl \'https://www.airkorea.or.kr/web/stationInfo/getRealExcel\' \
      -H \'Connection: keep-alive\' \
      -H \'Cache-Control: max-age=0\' \
      -H \'Upgrade-Insecure-Requests: 1\' \
      -H \'Origin: https://www.airkorea.or.kr\' \
      -H \'Content-Type: application/x-www-form-urlencoded\' \
      -H \'User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36\' \
      -H \'Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9\' \
      -H \'Sec-Fetch-Site: same-origin\' \
      -H \'Sec-Fetch-Mode: navigate\' \
      -H \'Sec-Fetch-User: ?1\' \
      -H \'Sec-Fetch-Dest: document\' \
      -H \'Referer: https://www.airkorea.o\r.kr/web/stationInfo?pMENU_NO=93\' \
      -H \'Accept-Language: en-US,en;q=0.9\' \
      -H \'Cookie: WMONID=D2jZJJF0hpI; tm_x=197923.160699; tm_y=451713.403762; station=111121; JSESSIONID=8iFAdhxmUSZF7ULqZuuRXdbjtRM2RNRlvbAsdxEt39N9GP6v44NVvyQmY1u9TxrA.airwas1_servlet_newairkorea4\' \
      --data-raw \'action=station&districtnum=&stationCode=&numValue=&loading=yes&leftShow=realTime&pMENU_NO=93&mang_code=3&district=\' \
      --compressed \
      --output \'{directory}kr_monitoring_site_locations_raw.xlsx\'')

    return None

def geocode_kr_monitor_locations():
    directory = datadir + 'korea/'
    monitor_locations = pd.read_excel(directory + 'kr_monitoring_site_locations_raw.xlsx', 
                                  skiprows=3, encoding ='cp949')
    monitor_locations.loc[:,'latitude'] = 0
    monitor_locations.loc[:,'longitude'] = 0
    apikey = pd.read_csv('apikey.csv').iloc[0,0]
    url = 'https://maps.googleapis.com/maps/api/geocode/json'
    for idx in monitor_locations.index:
        add = monitor_locations.loc[idx,'측정소 주소']
        params = {'address': add,'key':apikey}
        response = requests.get(url,params=params)
        resp_json_payload = response.json()
        if resp_json_payload['status'] == 'OK':
            monitor_locations.loc[idx,'latitude'] = pd.json_normalize(resp_json_payload['results'], max_level=3).loc[0,'geometry.location.lat']
            monitor_locations.loc[idx,'longitude'] = pd.json_normalize(resp_json_payload['results'], max_level=3).loc[0,'geometry.location.lng']
    monitor_locations.to_excel(directory + 'kr_monitoring_site_locations.xlsx')
    return None

def eu_monitor_utility():
    directory = datadir + 'europe/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    monloc = pd.read_csv('https://discomap.eea.europa.eu/map/fme/metadata/PanEuropean_metadata.csv',
           sep='\t')
    monloc.to_feather(directory + 'eu_monitoring_site_locations.feather')
    return None

def clean_us_monitor_file(mons_us):
    '''

    '''
    mons_us.drop('index',axis=1,inplace=True)
    def get_full_name(x):
        try:
            return us_abbrev_state[x]
        except:
            return 'Unknown'
    mons_us.countyname = mons_us.countyname.apply(get_full_name)
    tmp = dict.fromkeys(['NO2', 'PM2.5', 'O3'])
    output_cols = ['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude']
    relevant_cols = ['AQSID', 'parametername', 'Country', 'countyname', 'latitude',
                               'longitude']
    mons_us.loc[:, 'Country'] = 'United States'
    for k in ['NO2', 'PM2.5', 'O3']:
        idx = mons_us.parametername == k
        tmp[k] = mons_us.loc[idx, relevant_cols].rename(columns={i: j for i,j in zip(relevant_cols, output_cols)})
        tmp[k].loc[tmp[k].Region != 'Unknown', 'Country'] = 'United States'
    return tmp

def clean_jp_monitor_file(jp_monitors):
    ''' 

    '''
    jp_monitors = jp_monitors.iloc[:,[1,8,9,10,11,12,13,14,18]]
    lstcols = ['Station number','Station name','latitude_degree',
               'latitude_minute', 'latitude_Second', 'Longitude_Degree', 
               'Longitude_Minute','Longitude_Second',
              'Prefecture']
    jp_monitors = jp_monitors.rename(columns={col: lstcols[i] for i,col in enumerate(list(jp_monitors))})
    jp_monitors['latitude'] = jp_monitors['latitude_degree'] + jp_monitors['latitude_minute']/60 + jp_monitors['latitude_Second']/3600
    jp_monitors['longitude'] = jp_monitors['Longitude_Degree'] + jp_monitors['Longitude_Minute']/60 + jp_monitors['Longitude_Second']/3600
    jp_monitors = jp_monitors.drop(columns=['latitude_degree','latitude_minute', 'latitude_Second',
                           'Longitude_Degree', 'Longitude_Minute', 'Longitude_Second'])
    def normal_pref(pref):
        if pref[-2:] == 'do':
            return pref[:-3] + pref[-2:]
        elif pref[-2:] == 'to' or pref[-2:] == 'fu':
            return pref[:-3]
        elif pref[-3:] == 'ken':
            return pref[:-4]
    jp_monitors['Prefecture'] = jp_monitors['Prefecture'].apply(normal_pref)
    relevant_cols = ['Station number', 'Parameter', 'Country', 'Prefecture', 
                     'latitude', 'longitude']
    tmp = dict.fromkeys(['NO2', 'PM2.5', 'O3'])
    output_cols = ['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude']
    for k in ['NO2', 'PM2.5', 'O3']:
        jp_monitors.loc[:, 'Parameter'] = k
        jp_monitors.loc[:, 'Country'] = 'Japan'
        tmp[k] = jp_monitors[relevant_cols].rename(columns={i: j for i,j in zip(relevant_cols, output_cols)})
    return tmp

def clean_kr_monitor_file(kr_monitors):
    ''' 

    '''
    relevant_cols = ['Station number', 'Parameter', 'Country', 'Region', 
                     'latitude', 'longitude']
    tmp = dict.fromkeys(['NO2', 'PM2.5', 'O3'])
    output_cols = ['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude']
    for k in ['NO2', 'PM2.5', 'O3']:
        kr_monitors.loc[:, 'Parameter'] = k
        kr_monitors.loc[:, 'Country'] = 'South Korea'
        kr_monitors.loc[:, 'Region'] = 'Unknown'
        tmp[k] = kr_monitors[relevant_cols].rename(columns={i: j for i,j in zip(relevant_cols, output_cols)})
    return tmp

def clean_eu_monitor_file(eu_monitors):
    ''' 

    '''
    relevant_cols = ['AirQualityStation', 'Parameter', 'Country', 'Region', 
                 'Latitude', 'Longitude']
    tmp = dict.fromkeys(['NO2', 'PM2.5', 'O3'])
    output_cols = ['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude']
    def get_country(code):
        try: 
            return pycountry.countries.get(alpha_2=code).name
        except:
            return 'Unknown'
    for k in ['NO2', 'PM2.5', 'O3']:
        eu_monitors.loc[:, 'Parameter'] = k
        eu_monitors.loc[:, 'Country'] = eu_monitors.Countrycode.apply(get_country)
        eu_monitors.loc[:, 'Region'] = 'Unknown'
        tmp[k] = eu_monitors[relevant_cols].rename(columns={i: j for i,j in zip(relevant_cols, output_cols)})
    return tmp

def clean_ch_monitor_file(ch_monitors):
    ''' 

    '''
    relevant_cols = ['监测点编码', 'Parameter', 'Country', 'Region', 'latitude',
                'longitude']
    tmp = dict.fromkeys(['NO2', 'PM2.5', 'O3'])
    output_cols = ['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude']
    for k in ['NO2', 'PM2.5', 'O3']:
        ch_monitors.loc[:, 'Parameter'] = k
        ch_monitors.loc[:, 'Country'] = 'China'
        ch_monitors.loc[:, 'Region'] = 'Unknown'
        tmp[k] = ch_monitors[relevant_cols].rename(columns={i: j for i,j in zip(relevant_cols, output_cols)})
    return tmp

def load_raw_monitor_file(region):
    '''
    region: str containing 'US', 'Europe', 'China', 'Japan', or 'South Korea'
    '''
    if region == 'US':
        return pd.read_feather('./raw_data/us/us_monitoring_site_locations.feather')
    elif region == 'China':
        return pd.read_feather('./ext_data/ch_site_loc.feather')
    elif region == 'Japan':
        return pd.read_csv('./raw_data/japan/TM20170000.txt', 
            header=0, encoding='gb18030')
    elif region == 'South Korea':
        return pd.read_feather('./ext_data/kr_site_loc.feather')
    elif region == 'Europe':
        return pd.read_feather('./raw_data/europe/eu_monitoring_site_locations.feather')
    else:
        raise ValueError('Invalid region')

def select_clean_function(region):
    '''
    region: str containing 'US', 'Europe', 'China', 'Japan', or 'South Korea'
    '''
    if region == 'US':
        return clean_us_monitor_file
    elif region == 'China':
        return clean_ch_monitor_file
    elif region == 'Japan':
        return clean_jp_monitor_file
    elif region == 'South Korea':
        return clean_kr_monitor_file
    elif region == 'Europe':
        return clean_eu_monitor_file
    else:
        raise ValueError('Invalid region')

def merge_all_monitor_data():
    '''

    '''
    # Initialize
    monitors = dict.fromkeys(['NO2', 'O3', 'PM2.5'])
    for k in monitors.keys():
        monitors[k] = pd.DataFrame([], columns=['Identifier', 'Parameter', 'Country',
                                       'Region', 'Latitude', 'Longitude'])
        for region in ['US', 'Europe', 'China', 'South Korea', 'Japan']:
            mons = load_raw_monitor_file(region)
            clean_fn = select_clean_function(region)
            clean_df = clean_fn(mons)
            clean_df[k].drop_duplicates(inplace=True)
            monitors[k] = pd.concat([monitors[k], clean_df[k]], axis=0)
        # Clean-up
        monitors[k].reset_index(drop=True, inplace=True)
        monitors[k].Identifier = monitors[k].Identifier.astype(str)
        monitors[k] = monitors[k].loc[monitors[k].Latitude.notnull()]
        monitors[k] = monitors[k].loc[monitors[k].Longitude.notnull()]
        # Get population and regions
        monitors[k] = get_population(monitors[k])
        monitors[k] = monitors[k].loc[monitors[k].Population > 0]
        monitors[k] = get_region(monitors[k])
        monitors[k].reset_index(drop=True).to_feather(datadir + f'{k}_monitor_locations.feather')
    return monitors

def get_population(monitors):
    '''
    monitors: DataFrame with 6 columns: ID, Parameter, Country, Region, Latitude, Longitude
    '''
    steplat = steplon = 1.0/24
    lats = np.arange(-90+steplat/2, 90, steplat)
    lons = np.arange(-180+steplon/2, 180, steplon)
    # Find each monitor's grid cell
    monitors.loc[:, 'ilat'] = monitors.Latitude.apply(
        lambda x: int(np.floor((x - lats[0]) / steplat)))
    monitors.loc[:, 'ilon'] = monitors.Longitude.apply(
        lambda x: int(np.floor((x - lons[0]) / steplon)))
    # Get population
    pop = xr.open_dataset(
    './ext_data/gpw_v4_population_density_adjusted_rev11_2pt5_min.nc')
    var = 'UN WPP-Adjusted Population Density, v4.11 (2000, 2005, 2010, 2015, 2020): 2.5 arc-minutes'
    pop15 = pop[var].isel(raster=3) * pop[var].isel(raster=8) # Density * surface area
    pop15.values = np.flipud(pop15.values)
    pop15['latitude'].values = pop15['latitude'].values[::-1]
    pop15.values[pd.isnull(pop15.values)] = 0
    monitors.loc[:, 'Population'] = 0
    for idx in monitors.index:
        monitors.loc[idx, 'Population'] = pop15.values[monitors.loc[idx,'ilat'], 
                                                      monitors.loc[idx, 'ilon']]
    # Remove double counting
    tmp = monitors.groupby(['ilat', 'ilon'])['Population'].count()
    for dupix in tmp.index:
        idx = (monitors['ilat'] == dupix[0]).values * \
            (monitors['ilon'] == dupix[1]).values
        monitors.loc[idx, 'Population'] /= tmp.loc[dupix[0], dupix[1]]
    monitors.drop(['ilat', 'ilon'], axis=1, inplace=True)
    return monitors

def get_region(monitors):
    '''
    monitors: DataFrame with 6 columns: ID, Parameter, Country, Region, Latitude, Longitude
    '''
    geom = [Point(xy) for xy in zip(monitors.Longitude, monitors.Latitude)]
    monitors = gpd.GeoDataFrame(monitors, crs='EPSG:4326', geometry=geom)
    monitors = gpd.sjoin(monitors.drop(['Country', 'Region'], axis=1), 
                         all_regions[['Country', 'Region', 'geometry']], op='within')
    monitors.drop(['geometry', 'index_right'], axis=1, inplace=True)
    return monitors


